#define _XOPEN_SOURCE 700
#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <limits.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include "linkify.h"
#include "str.h"
#include "unicode.h"
#include "util.h"
#include "style.h"

char *strstr(const char *haystack, const char *needle);
char *strerror(int errnum);

static struct str *parse_section(struct parser *p) {
	struct str *section = str_create();
	uint32_t ch;
	char *subsection;
	while ((ch = parser_getch(p)) != UTF8_INVALID) {
		if (ch < 0x80 && isalnum((unsigned char)ch)) {
			int ret = str_append_ch(section, ch);
			assert(ret != -1);
		} else if (ch == ')') {
			if (section->len == 0) {
				break;
			}
			int sec = strtol(section->str, &subsection, 10);
			if (section->str == subsection) {
				parser_fatal(p, "Expected section digit");
				break;
			}
			if (sec < 0 || sec > 9) {
				parser_fatal(p, "Expected section between 0 and 9");
				break;
			}
			return section;
		} else {
			parser_fatal(p, "Expected alphanumerical character or )");
			break;
		}
	};
	parser_fatal(p, "Expected manual section");
	return NULL;
}

static struct str *parse_extra(struct parser *p) {
	struct str *extra = str_create();
	int ret = str_append_ch(extra, '"');
	assert(ret != -1);
	uint32_t ch;
	while ((ch = parser_getch(p)) != UTF8_INVALID) {
		if (ch == '"') {
			ret = str_append_ch(extra, ch);
			assert(ret != -1);
			return extra;
		} else if (ch == '\n') {
			parser_fatal(p, "Unclosed extra preamble field");
			break;
		} else {
			ret = str_append_ch(extra, ch);
			assert(ret != -1);
		}
	}
	str_free(extra);
	return NULL;
}

static void parse_preamble(struct parser *p, bool fragment) {
	struct str *name = str_create();
	int ex = 0;
	struct str *extras[2] = { NULL };
	struct str *section = NULL;
	uint32_t ch;
	time_t date_time;
	char date[256];
	char *source_date_epoch = getenv("SOURCE_DATE_EPOCH");
	if (source_date_epoch != NULL) {
		unsigned long long epoch;
		char *endptr;
		errno = 0;
		epoch = strtoull(source_date_epoch, &endptr, 10);
		if ((errno == ERANGE && (epoch == ULLONG_MAX || epoch == 0))
				|| (errno != 0 && epoch == 0)) {
			fprintf(stderr, "$SOURCE_DATE_EPOCH: strtoull: %s\n",
					strerror(errno));
			exit(EXIT_FAILURE);
		}
		if (endptr == source_date_epoch) {
			fprintf(stderr, "$SOURCE_DATE_EPOCH: No digits were found: %s\n",
					endptr);
			exit(EXIT_FAILURE);
		}
		if (*endptr != '\0') {
			fprintf(stderr, "$SOURCE_DATE_EPOCH: Trailing garbage: %s\n",
					endptr);
			exit(EXIT_FAILURE);
		}
		if (epoch > ULONG_MAX) {
			fprintf(stderr, "$SOURCE_DATE_EPOCH: value must be smaller than or "
					"equal to %lu but was found to be: %llu \n",
					ULONG_MAX, epoch);
			exit(EXIT_FAILURE);
		}
		date_time = epoch;
	} else {
		date_time = time(NULL);
	}
	struct tm *date_tm = gmtime(&date_time);
	strftime(date, sizeof(date), "%F", date_tm);
	while ((ch = parser_getch(p)) != UTF8_INVALID) {
		if ((ch < 0x80 && isalnum((unsigned char)ch))
				|| ch == '_' || ch == '-' || ch == '.') {
			int ret = str_append_ch(name, ch);
			assert(ret != -1);
		} else if (ch == '(') {
			section = parse_section(p);
		} else if (ch == '"') {
			if (ex == 2) {
				parser_fatal(p, "Too many extra preamble fields");
			}
			extras[ex++] = parse_extra(p);
		} else if (ch == '\n') {
			if (name->len == 0) {
				parser_fatal(p, "Expected preamble");
			}
			if (section == NULL) {
				parser_fatal(p, "Expected manual section");
			}
			if (!fragment) {
				fprintf(p->output, style_header, name->str, section->str);
			}
			fprintf(p->output, "<header>\n");
			fprintf(p->output, "<h1>%s(%s)</h1>\n", name->str, section->str);
			fprintf(p->output, "<p class=\"header-date\">%s</p>\n", date);
			fprintf(p->output, "</header>\n");
			/* ex2 and ex3 are already double-quoted */
			// TODO use ex2 and ex3 if present
			//if (ex2) {
			//	fprintf(p->output, " %s", ex2);
			//}
			//if (ex3) {
			//	fprintf(p->output, " %s", ex3);
			//}
			//fprintf(p->output, "\n");
			break;
		} else if (section == NULL) {
			parser_fatal(p, "Name characters must be A-Z, a-z, 0-9, `-`, `_`, or `.`");
		}
	}
	str_free(name);
	for (int i = 0; i < 2; ++i) {
		if (extras[i] != NULL) {
			str_free(extras[i]);
		}
	}
}

static void parse_format(struct parser *p, enum formatting fmt) {
	char formats[FORMAT_LAST] = {
		[FORMAT_BOLD] = 'b',
		[FORMAT_UNDERLINE] = 'u',
	};
	char error[512];
	if (p->flags) {
		if ((p->flags & ~fmt)) {
			snprintf(error, sizeof(error), "Cannot nest inline formatting "
						"(began with %c at %d:%d)",
					p->flags == FORMAT_BOLD ? '*' : '_',
					p->fmt_line, p->fmt_col);
			parser_fatal(p, error);
		}
		fprintf(p->output, "</%c>", formats[p->flags]);
	} else {
		fprintf(p->output, "<%c>", formats[fmt]);
		p->fmt_line = p->line;
		p->fmt_col = p->col;
	}
	p->flags ^= fmt;
}

static bool parse_linebreak(struct parser *p) {
	uint32_t plus = parser_getch(p);
	if (plus != '+') {
		fprintf(p->output, "+");
		parser_pushch(p, plus);
		return false;
	}
	uint32_t lf = parser_getch(p);
	if (lf != '\n') {
		fprintf(p->output, "+");
		parser_pushch(p, lf);
		parser_pushch(p, plus);
		return false;
	}
	uint32_t ch = parser_getch(p);
	if (ch == '\n') {
		parser_fatal(
				p, "Explicit line breaks cannot be followed by a blank line");
	}
	parser_pushch(p, ch);
	fprintf(p->output, "<br>\n");
	return true;
}

static void parse_text(struct parser *p, bool linkify);

static uint32_t parse_word(struct parser *p) {
#define WBSZ 256
	uint32_t ch = 0;
	int i = 0;
	static unsigned char wordbuf[WBSZ];
	memset(wordbuf, 0, WBSZ);
	FILE *output = fmemopen(wordbuf, WBSZ, "w");
	for (i = 0; i < WBSZ - 4;) {
		ch = parser_getch(p);
		if (ch == UTF8_INVALID) {
			break;
		}
		i += utf8_fputch(output, ch);
		if (ch == '>' || ch == '\n' || ch == ' ') {
			break;
		}
	}
	fclose(output);
	uint8_t *remainder = wordbuf;
	// Try to linkify if more than 3 characters and word ended by final
	// character, not by buffer length.
	if (i > 3 && (ch == UTF8_INVALID || ch == '>' || ch == '\n' || ch == ' ')) {
		try_linkify(p->output, wordbuf, (const uint8_t **)(&remainder));
	}
	if ((remainder - wordbuf) < i) {
		// Not linkified, or some remainder, process as regular text.
		// Save state
		FILE *input = p->input;
		const char *str = p->str;
		// process
		p->input = fmemopen(remainder, i - (remainder - wordbuf), "rw");
		p->str = 0;
		parse_text(p, false);
		// restore state
		p->input = input;
		p->str = str;
	}
	return ch;
}

static void parse_text(struct parser *p, bool linkify) {
	uint32_t ch, next, last = ' ';
	int i = 0;
	while ((ch = parser_getch(p)) != UTF8_INVALID) {
		switch (ch) {
		case '\\':
			ch = parser_getch(p);
			if (ch == UTF8_INVALID) {
				parser_fatal(p, "Unexpected EOF");
			} else {
				utf8_fputch(p->output, ch);
			}
			break;
		case '*':
			parse_format(p, FORMAT_BOLD);
			last = ch;
			break;
		case '_':
			next = parser_getch(p);
			if (!isalnum((unsigned char)last) || (
						(p->flags & FORMAT_UNDERLINE) &&
						!isalnum((unsigned char)next))) {
				parse_format(p, FORMAT_UNDERLINE);
			} else {
				utf8_fputch(p->output, ch);
			}
			if (next == UTF8_INVALID) {
				return;
			}
			parser_pushch(p, next);
			last = ch;
			break;
		case '+':
			if (parse_linebreak(p)) {
				last = '\n';
			}
			break;
		case '\n':
			utf8_fputch(p->output, ch);
			return;
		case '<':
			last = ch;
			fprintf(p->output, "&lt;");
			if (linkify) {
				last = parse_word(p);
				if (last == '\n' || last == UTF8_INVALID) {
					return;
				}
				break;
			}
			break;
		case '>':
			last = ch;
			fprintf(p->output, "&gt;");
			break;
		case '.':
		case '\'':
		case '!':
		case '?':
			last = ch;
			utf8_fputch(p->output, ch);
			break;
		default:
			if (linkify && last == ' ') {
				parser_pushch(p, ch);
				last = parse_word(p);
				if (last == '\n' || last == UTF8_INVALID) {
					return;
				}
				break;
			}
			last = ch;
			utf8_fputch(p->output, ch);
			break;
		}
		++i;
	}
}

static void parse_heading(struct parser *p) {
	uint32_t ch;
	int level = 1;
	while ((ch = parser_getch(p)) != UTF8_INVALID) {
		if (ch == '#') {
			++level;
		} else if (ch == ' ') {
			break;
		} else {
			parser_fatal(p, "Invalid start of heading (probably needs a space)");
		}
	}
	if (level > 2) {
		parser_fatal(p, "Only headings up to two levels deep are permitted");
	}
	level++;

	struct str *heading = str_create();
	while ((ch = parser_getch(p)) != UTF8_INVALID) {
		if (ch == '\n') {
			break;
		}
		str_append_ch(heading, ch);
	}

	fprintf(p->output, "<h%d id=\"", level);
	html_id(p, heading->str);
	fprintf(p->output, "\">");
	html_escape(p, heading->str);
	fprintf(p->output, " <a aria-hidden=\"true\" href=\"#");
	html_id_fragment(p, heading->str);
	fprintf(p->output, "\">¶</a></h%d>\n", level);
}

static int parse_indent(struct parser *p, int *indent, bool write) {
	int i = 0;
	uint32_t ch;
	while ((ch = parser_getch(p)) == '\t') {
		++i;
	}
	parser_pushch(p, ch);
	if ((ch == '\n' || ch == UTF8_INVALID) && *indent != 0) {
		// Don't change indent when we encounter empty lines or EOF
		return *indent;
	}
	if (write) {
		if ((i - *indent) > 1) {
			parser_fatal(p, "Indented by an amount greater than 1");
		} else if (i < *indent) {
			end_paragraph(p);
			start_paragraph(p, i);
		} else if (i == *indent + 1) {
			end_paragraph(p);
			start_paragraph(p, i);
		}
	}
	*indent = i;
	return i;
}

static void list_header(struct parser *p, int *num) {
	fprintf(p->output, "<li>");
	if (*num != -1) {
		*num = *num + 1;
	}
}

static void parse_list(struct parser *p, int *indent, int num) {
	uint32_t ch;
	if ((ch = parser_getch(p)) != ' ') {
		parser_fatal(p, "Expected space before start of list entry");
	}
	if (num == -1) {
		fprintf(p->output, "<ul>\n");
	} else {
		fprintf(p->output, "<ol>\n");
	}

	int initial_indent = *indent;
	list_header(p, &num);
	parse_text(p, true);
	do {
		int cur = *indent;
		parse_indent(p, indent, false);
		if ((ch = parser_getch(p)) == UTF8_INVALID) {
			break;
		}
		switch (ch) {
		case ' ':
			if ((ch = parser_getch(p)) != ' ') {
				parser_fatal(p, "Expected two spaces for list entry continuation");
			}
			parse_text(p, true);
			break;
		case '-':
		case '.':
			if (*indent > cur) {
				fprintf(p->output, "<%s>\n", num == -1 ? "ul" : "ol");
			} else {
				if (*indent < cur) {
					fprintf(p->output, "</li></%s>\n", num == -1 ? "ul" : "ol");
				}
				fprintf(p->output, "</li>\n");
			}
			if ((ch = parser_getch(p)) != ' ') {
				parser_fatal(p, "Expected space before start of list entry");
			}
			list_header(p, &num);
			parse_text(p, true);
			break;
		default:
			for (cur = *indent; cur >= initial_indent; cur--) {
				fprintf(p->output, "</%s>\n", num == -1 ? "ul" : "ol");
			}
			return;
		}
	} while (ch != UTF8_INVALID);
}

static void parse_literal(struct parser *p, int *indent) {
	uint32_t ch;
	if ((ch = parser_getch(p)) != '`' ||
		(ch = parser_getch(p)) != '`' ||
		(ch = parser_getch(p)) != '\n') {
		parser_fatal(p, "Expected ``` and a newline to begin literal block");
	}
	int stops = 0;
	fprintf(p->output, "<pre><code>\n");
	bool check_indent = true;
	do {
		if (check_indent) {
			int _indent = *indent;
			parse_indent(p, &_indent, false);
			if (_indent < *indent) {
				parser_fatal(p, "Cannot deindent in literal block");
			}
			while (_indent > *indent) {
				--_indent;
				fprintf(p->output, "\t");
			}
			check_indent = false;
		}
		if ((ch = parser_getch(p)) == UTF8_INVALID) {
			break;
		}
		if (ch == '`') {
			if (++stops == 3) {
				if ((ch = parser_getch(p)) != '\n') {
					parser_fatal(p, "Expected literal block to end with newline");
				}
				fprintf(p->output, "\n</code></pre>\n");
				return;
			}
		} else {
			while (stops != 0) {
				fputc('`', p->output);
				--stops;
			}
			switch (ch) {
			case '\\':
				ch = parser_getch(p);
				if (ch == UTF8_INVALID) {
					parser_fatal(p, "Unexpected EOF");
				} else if (ch == '\\') {
					fprintf(p->output, "\\\\");
				} else {
					utf8_fputch(p->output, ch);
				}
				break;
			case '&':
				fprintf(p->output, "&amp;");
				break;
			case '<':
				fprintf(p->output, "&lt;");
				break;
			case '>':
				fprintf(p->output, "&gt;");
				break;
			case '\n':
				check_indent = true;
				/* fallthrough */
			default:
				utf8_fputch(p->output, ch);
				break;
			}
		}
	} while (ch != UTF8_INVALID);
}

enum table_align {
	ALIGN_LEFT,
	ALIGN_CENTER,
	ALIGN_RIGHT,
	ALIGN_LEFT_EXPAND,
	ALIGN_CENTER_EXPAND,
	ALIGN_RIGHT_EXPAND,
};

struct table_row {
	struct table_cell *cell;
	struct table_row *next;
};

struct table_cell {
	enum table_align align;
	struct str *contents;
	struct table_cell *next;
};

static void parse_table(struct parser *p, uint32_t style) {
	struct table_row *table = NULL;
	struct table_row *currow = NULL, *prevrow = NULL;
	struct table_cell *curcell = NULL;
	int column = 0;
	int numcolumns = -1;
	uint32_t ch;
	parser_pushch(p, '|');

	do {
		if ((ch = parser_getch(p)) == UTF8_INVALID) {
			break;
		}
		switch (ch) {
		case '\n':
			goto commit_table;
		case '|':
			prevrow = currow;
			currow = xcalloc(1, sizeof(struct table_row));
			if (prevrow) {
				if (column != numcolumns && numcolumns != -1) {
					parser_fatal(p, "Each row must have the "
							"same number of columns");
				}
				numcolumns = column;
				column = 0;
				prevrow->next = currow;
			}
			curcell = xcalloc(1, sizeof(struct table_cell));
			currow->cell = curcell;
			if (!table) {
				table = currow;
			}
			break;
		case ':':
			if (!currow) {
				parser_fatal(p, "Cannot start a column without "
						"starting a row first");
			} else {
				struct table_cell *prev = curcell;
				curcell = xcalloc(1, sizeof(struct table_cell));
				if (prev) {
					prev->next = curcell;
				}
				++column;
			}
			break;
		case ' ':
			goto continue_cell;
		default:
			parser_fatal(p, "Expected either '|' or ':'");
			break;
		}
		if ((ch = parser_getch(p)) == UTF8_INVALID) {
			break;
		}
		switch (ch) {
		case '[':
			curcell->align = ALIGN_LEFT;
			break;
		case '-':
			curcell->align = ALIGN_CENTER;
			break;
		case ']':
			curcell->align = ALIGN_RIGHT;
			break;
		case '<':
			curcell->align = ALIGN_LEFT_EXPAND;
			break;
		case '=':
			curcell->align = ALIGN_CENTER_EXPAND;
			break;
		case '>':
			curcell->align = ALIGN_RIGHT_EXPAND;
			break;
		case ' ':
			if (prevrow) {
				struct table_cell *pcell = prevrow->cell;
				for (int i = 0; i <= column && pcell; ++i, pcell = pcell->next) {
					if (i == column) {
						curcell->align = pcell->align;
						break;
					}
				}
			} else {
				parser_fatal(p, "No previous row to infer alignment from");
			}
			break;
		default:
			parser_fatal(p, "Expected one of '[', '-', ']', or ' '");
			break;
		}
		curcell->contents = str_create();
continue_cell:
		switch (ch = parser_getch(p)) {
		case ' ':
			// Read out remainder of the text
			while ((ch = parser_getch(p)) != UTF8_INVALID) {
				switch (ch) {
				case '\n':
					goto commit_cell;
				default:;
					int ret = str_append_ch(curcell->contents, ch);
					assert(ret != -1);
					break;
				}
			}
			break;
		case '\n':
			goto commit_cell;
		default:
			parser_fatal(p, "Expected ' ' or a newline");
			break;
		}
commit_cell:
		if (strstr(curcell->contents->str, "T{")
				|| strstr(curcell->contents->str, "T}")) {
			parser_fatal(p, "Cells cannot contain T{ or T} "
					"due to roff limitations");
		}
	} while (ch != UTF8_INVALID);
commit_table:

	if (ch == UTF8_INVALID) {
		return;
	}

	const char *expand = "";

	// Figure out alignments first
	currow = table;
	while (currow) {
		curcell = currow->cell;
		while (curcell) {
			switch (curcell->align) {
			case ALIGN_LEFT_EXPAND:
			case ALIGN_CENTER_EXPAND:
			case ALIGN_RIGHT_EXPAND:
				expand = " expand";
				break;
			default:
				break;
			}
			curcell = curcell->next;
		}
		currow = currow->next;
	}

	const char *box = "";

	switch (style) {
	case '[':
		box = "allbox";
		break;
	case ']':
		box = "box";
		break;
	default:
		box = "nobox";
		break;
	}

	fprintf(p->output, "<table class=\"%s%s\"><tbody>\n", box, expand);
	p->p_open = 1;

	// Then contents
	currow = table;
	while (currow) {
		fprintf(p->output, "<tr class=\"%s\">\n", box);
		curcell = currow->cell;
		while (curcell) {
			const char *align = "";
			expand = "";
			switch (curcell->align) {
				case ALIGN_LEFT_EXPAND:
					expand = " expand";
					/* fallthrough */
				case ALIGN_LEFT:
					align = " left";
					break;
				case ALIGN_CENTER_EXPAND:
					expand = " expand";
					/* fallthrough */
				case ALIGN_CENTER:
					align = " center";
					break;
				case ALIGN_RIGHT_EXPAND:
					expand = " expand";
					/* fallthrough */
				case ALIGN_RIGHT:
					align = " right";
					break;
			}
			fprintf(p->output, "<td class=\"%s%s%s\">\n", box, align, expand);
			parser_pushstr(p, curcell->contents->str);
			parse_text(p, true);
			fprintf(p->output, "</td>\n");
			struct table_cell *prev = curcell;
			curcell = curcell->next;
			str_free(prev->contents);
			free(prev);
		}
		fprintf(p->output, "</tr>\n");
		struct table_row *prev = currow;
		currow = currow->next;
		free(prev);
	}

	fprintf(p->output, "</tbody></table>\n");
	p->p_open = 0;
}

static void parse_document(struct parser *p) {
	uint32_t ch;
	int indent = 0, list_indent = 0;
	do {
		parse_indent(p, &indent, true);
		if ((ch = parser_getch(p)) == UTF8_INVALID) {
			break;
		}
		switch (ch) {
		case ';':
			if ((ch = parser_getch(p)) != ' ') {
				parser_fatal(p, "Expected space after ; to begin comment");
			}
			do {
				ch = parser_getch(p);
			} while (ch != UTF8_INVALID && ch != '\n');
			break;
		case '#':
			if (indent != 0) {
				parser_pushch(p, ch);
				start_paragraph(p, indent);
				parse_text(p, true);
				break;
			}
			end_paragraph(p);
			if ((ch = parser_getch(p)) == '#') {
				parser_pushch(p, ch);
				start_subsection(p);
			} else {
				parser_pushch(p, ch);
				start_section(p);
			}
			parse_heading(p);
			break;
		case '-':
			list_indent = indent;
			parse_list(p, &list_indent, -1);
			break;
		case '.':
			if ((ch = parser_getch(p)) == ' ') {
				parser_pushch(p, ch);
				parse_list(p, &indent, 1);
			} else {
				parser_pushch(p, ch);
				start_paragraph(p, indent);
				parse_text(p, true);
			}
			break;
		case '`':
			parse_literal(p, &indent);
			break;
		case '[':
		case '|':
		case ']':
			if (indent != 0) {
				parser_fatal(p, "Tables cannot be indented");
			}
			end_paragraph(p);
			parse_table(p, ch);
			break;
		case ' ':
			parser_fatal(p, "Tabs are required for indentation");
			break;
		case '\n':
			if (p->flags) {
				char error[512];
				snprintf(error, sizeof(error), "Expected %c before starting "
						"new paragraph (began with %c at %d:%d)",
						p->flags == FORMAT_BOLD ? '*' : '_',
						p->flags == FORMAT_BOLD ? '*' : '_',
						p->fmt_line, p->fmt_col);
				parser_fatal(p, error);
			}
			end_paragraph(p);
			break;
		default:
			parser_pushch(p, ch);
			start_paragraph(p, indent);
			parse_text(p, true);
			break;
		}
	} while (ch != UTF8_INVALID);
	end_paragraph(p);
	end_section(p);
}

int main(int argc, char **argv) {
	bool fragment = false;

	if (argc == 2 && strcmp(argv[1], "-v") == 0) {
		printf("scd2html " VERSION "\n");
		return 0;
	} else if (argc == 2 && strcmp(argv[1], "-f") == 0) {
		fragment = true;
	} else if (argc > 1) {
		fprintf(stderr, "Usage: scd2html [ -f ] < input.scd > output.html\n");
		return 1;
	}
	struct parser p = {
		.input = stdin,
		.output = stdout,
		.line = 1,
		.col = 1,
		.p_open = 0
	};
	parse_preamble(&p, fragment);
	parse_document(&p);
	if (!fragment) {
		fprintf(p.output, style_footer);
	}
	return 0;
}
