#include <stdarg.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include "unicode.h"
#include "util.h"

void parser_fatal(struct parser *parser, const char *err) {
	fprintf(stderr, "Error at %d:%d: %s\n",
			parser->line, parser->col, err);
	fclose(parser->input);
	fclose(parser->output);
	exit(1);
}

uint32_t parser_getch(struct parser *parser) {
	if (parser->qhead) {
		return parser->queue[--parser->qhead];
	}
	if (parser->str) {
		uint32_t ch = utf8_decode(&parser->str);
		if (!ch || ch == UTF8_INVALID) {
			parser->str = NULL;
			return UTF8_INVALID;
		}
		return ch;
	}
	uint32_t ch = utf8_fgetch(parser->input);
	if (ch == '\n') {
		parser->col = 0;
		++parser->line;
	} else {
		++parser->col;
	}
	return ch;
}

void parser_pushch(struct parser *parser, uint32_t ch) {
	if (ch != UTF8_INVALID) {
		parser->queue[parser->qhead++] = ch;
	}
}

void parser_pushstr(struct parser *parser, const char *str) {
	if (parser->qhead) {
		parser_fatal(parser, "String pushed back to parser on non-empty queue");
	}
	parser->str = str;
}

int start_paragraph(struct parser *p, int indent) {
	if (p->p_open || p->t_open) {
		return 0;
	}
	if (indent) {
		p->p_open = indent + 1;
		for (; indent > 0; indent--) {
			fprintf(p->output, "\n<blockquote>");
		}
		return 1;
	} else {
		p->p_open = 1;
		return fprintf(p->output, "\n<p>");
	}
}

int end_paragraph(struct parser *p) {
	if (!p->p_open) {
		return 0;
	}
	if (p->p_open > 1) {
		for (int indent = p->p_open - 1; indent > 0; indent--) {
			fprintf(p->output, "</blockquote>\n");
		}
		p->p_open = 0;
		return 1;
	} else {
		p->p_open = 0;
		return fprintf(p->output, "</p>\n");
	}
}

void start_section(struct parser *p) {
	end_subsection(p);
	end_section(p);
	p->sh_open = 1;
	fprintf(p->output, "<section>\n");
}

void end_section(struct parser *p) {
	end_subsection(p);
	if (p->sh_open) {
		fprintf(p->output, "</section>\n");
		p->sh_open = 0;
	}
}

void start_subsection(struct parser *p) {
	end_subsection(p);
	p->ss_open = 1;
	fprintf(p->output, "<section>\n");
}

void end_subsection(struct parser *p) {
	if (p->ss_open) {
		fprintf(p->output, "</section>\n");
		p->ss_open = 0;
	}
}

void html_escape(struct parser *p, const char *value) {
	while (*value) {
		uint32_t ch = utf8_decode(&value);
		if (ch == '<') {
			fprintf(p->output, "&lt;");
		} else if (ch == '>') {
			fprintf(p->output, "&gt;");
		} else if (ch == '&') {
			fprintf(p->output, "&amp;");
		} else {
			utf8_fputch(p->output, ch);
		}
	}
}

void html_id(struct parser *p, const char *value) {
	while (*value) {
		uint32_t ch = utf8_decode(&value);
		if (ch == '<') {
			fprintf(p->output, "&lt;");
		} else if (ch == '>') {
			fprintf(p->output, "&gt;");
		} else if (ch == '"') {
			fprintf(p->output, "&quot;");
		} else if (ch == '&') {
			fprintf(p->output, "&amp;");
		} else if (ch == ' ') {
			fputc('_', p->output);
		} else {
			utf8_fputch(p->output, ch);
		}
	}
}

void html_id_fragment(struct parser *p, const char *value) {
	while (*value) {
		uint32_t ch = utf8_decode(&value);
		switch (ch) {
		case ' ':
			fputc('_', p->output);
			break;
		case '<':
		case '>':
		case '"':
		case '`':
		case '\'':
			fprintf(p->output, "%%%2X", ch);
			break;
		default:
			utf8_fputch(p->output, ch);
		}
	}
}

void *xcalloc(size_t n, size_t s) {
	void *p = calloc(n, s);
	if (!p) {
		fputs("Out of memory\n", stderr);
		abort();
	}
	return p;
}

void *xrealloc(void *p, size_t s) {
	void *ret = realloc(p, s);
	if (!ret) {
		fputs("Out of memory\n", stderr);
		abort();
	}
	return ret;
}
